#!/bin/python3
#%%
import pandas as pd

from tools.modeling.evaluation_tools import series_auc_and_ks_score
from tools.modeling.qcut_tools import append_intervel_col, get_distri_df, get_lift_df, get_psi_series, get_quantile_bins


#%%

def stat_y_by_month(
        data_df: pd.DataFrame,
        y_col: str = 'y',
        time_col: str = 'month',
        total = True,
        ):
    notna_idx = data_df[y_col].notna()
    y_df = data_df.loc[notna_idx, :]
    gp = y_df.groupby(by=time_col)[y_col]
    y_stat_df = pd.DataFrame({
        'count': gp.count(),
        'percent': gp.count() / len(y_df),
        'pos_cnt': gp.sum(),
        'pos_pct': gp.mean(),
    })
    if total:
        y_total_df = pd.DataFrame({
            'count': [y_stat_df['count'].sum()],
            'percent': [y_stat_df['percent'].sum()],
            'pos_cnt': [y_stat_df['pos_cnt'].sum()],
            'pos_pct': [y_stat_df['pos_cnt'].sum() / y_stat_df['count'].sum()],
        }, index=['total'])
        y_stat_df = pd.concat([y_stat_df, y_total_df], axis=0)
    y_stat_df.index.name = 'month'
    y_stat_df.reset_index(inplace = True)
    y_stat_df.index = pd.Index(range(1, len(y_stat_df)+1), name='index')
    return y_stat_df


def stat_sample_by_time(
        data_df: pd.DataFrame,
        time_col: str = 'month',
        total = True,
        ):
    gp = data_df.groupby(by=time_col)[time_col]
    stat_df = pd.DataFrame({
        'count': gp.count(),
        'percent': gp.count() / len(data_df),
    })
    if total:
        total_df = pd.DataFrame({
            'count': [stat_df['count'].sum()],
            'percent': [stat_df['percent'].sum()],
        }, index=['total'])
        stat_df = pd.concat([stat_df, total_df], axis=0)
    stat_df.index.name = 'month'
    stat_df.reset_index(inplace = True)
    stat_df.index = pd.Index(range(1, len(stat_df)+1), name='index')
    return stat_df


#%%

def get_auc_and_ks_by_group(
        data_df, 
        group_col: str,
        y_col: str = 'y',
        x_col: str = 'x',
        date_col: str = 'date',
        order = None,
        total = True,
        ):
    data_df = data_df.loc[
        data_df[y_col].notna(), 
        [date_col, group_col, y_col, x_col]
        ].copy()
    gp = data_df.groupby(by=group_col)
    date_stat_df = gp.apply(lambda df: pd.Series({
        'min_date': df[date_col].min(), 
        'max_date': df[date_col].max(),
        'count': df[group_col].count(),
        'percent': df[group_col].count() / len(data_df),
        'pos_cnt': df[y_col].sum(),
        'pos_pct': df[y_col].mean(),
        }))
    metric_stat_df = gp.apply(lambda df: series_auc_and_ks_score(
        df[y_col], df[x_col]
        ))

    stat_df = pd.concat([date_stat_df, metric_stat_df], axis=1)
    if order is None:
        order = sorted(data_df[group_col].unique())
    stat_df = stat_df.loc[order]

    if total:
        total_df = pd.DataFrame({
            'min_date': [data_df[date_col].min()],
            'max_date': [data_df[date_col].max()],
            'count': [len(data_df)],
            'percent': [1],
            'pos_cnt': [data_df[y_col].sum()],
            'pos_pct': [data_df[y_col].mean()],
            'auc': [series_auc_and_ks_score(data_df[y_col], data_df[x_col], metric='auc')],
            'ks': [series_auc_and_ks_score(data_df[y_col], data_df[x_col], metric='ks')],
        }, index=['total'])
        total_df.index.name = stat_df.index.name
        stat_df = pd.concat([stat_df, total_df], axis=0)

    stat_df.reset_index(inplace = True)
    stat_df.index = pd.Index(range(1, len(stat_df)+1), name='index')

    return stat_df


def get_auc_ks_coverage_by_group(
        y_df: pd.DataFrame,
        x_df: pd.DataFrame = None,
        y_cols: list = ['y'],
        x_cols: list = 'auto',
        ):
    if isinstance(y_df, pd.Series):
        y_df = y_df.to_frame()
    if isinstance(x_df, pd.Series):
        x_df = x_df.to_frame()
    if isinstance(y_cols, str):
        y_cols = [y_cols]
    if x_df is None:
        x_df = y_df
    if x_cols == 'auto':
        x_cols = x_df.select_dtypes(include='number').columns
    s = []
    for y_col in y_cols:
        for x_col in x_cols:
            d = {}
            auc_ks_s = series_auc_and_ks_score(
                y_df[y_col],
                x_df[x_col]
                )
            d['y_col'] = y_col
            d['x_col'] = x_col
            d[f'auc'] = auc_ks_s['auc']
            d[f'ks'] = auc_ks_s['ks']
            d['coverage'] = x_df[x_col].notna().mean()
            s.append(d)

    stat_df = pd.DataFrame(s)
    stat_df.set_index(y_cols + x_cols, inplace=True)

    return stat_df



#%%

def get_distri_by_group(
        data_df: pd.DataFrame,
        col_name: str = 'x',
        group_col: str = 'month',
        n: int = 20,
        drop_endpoints: bool = True,
        add_endpoints: list = [0, 100],
        interval_col: str = 'interval',
        reset_index: bool = True,
        psi_main_col = None,
        total = False,
    ):
    data_intervel_df = append_intervel_col(
        data_df=data_df, 
        col_name=col_name, 
        n=n, 
        drop_endpoints=drop_endpoints, 
        add_endpoints=add_endpoints
        )
    total_count_s = data_intervel_df.groupby(by=[interval_col], observed=False)[interval_col].count()
    count_s = data_intervel_df.groupby(by=[interval_col, group_col], observed=False)[interval_col].count()
    count_df = count_s.unstack(level=group_col)
    count_df['total'] = total_count_s
    percent_df = count_df / count_df.sum()

    if psi_main_col is not None:
        psi_df = pd.DataFrame({
            col: get_psi_series(count_df[col], count_df[psi_main_col], return_type='series') for col in percent_df.columns
        })
    else:
        psi_df = pd.DataFrame()
    stat_df = pd.concat([
        count_df.add_suffix('_count'),
        percent_df.add_suffix('_percent'), 
        psi_df.add_suffix('_psi'),
        ], axis=1)
    stat_df.columns.name = None

    if reset_index:
        stat_df.reset_index(inplace=True)
        stat_df.index = pd.Index(range(1, len(stat_df)+1), name='index')

    return stat_df


#%%

def get_lift_by_month(
        data_df: pd.DataFrame,
        x_col = 'x',
        n = 20,
        drop_endpoints=True,
        add_endpoints=[0, 100],
        group_col = 'month',
        ):

    quantiles = get_quantile_bins(
        data_df[x_col],
        n = n,
        drop_endpoints=drop_endpoints,
        add_endpoints=add_endpoints,
        )

    total_lift_df = get_lift_df(
        data_df=data_df, quantiles=quantiles, reset_index=False, NaN=False,
        ).add_prefix('total_')

    group_lift_df = data_df.groupby(by=group_col).apply(
        lambda df: get_lift_df(
            data_df=df, quantiles=quantiles, reset_index=False, NaN=False,
            ))
    group_lift_df = group_lift_df.unstack(level=group_col)
    group_lift_df.columns = [f'{p}_{c}' for c, p in group_lift_df.columns]

    lift_df = pd.concat([total_lift_df, group_lift_df], axis=1)
    
    lift_df.index.name = 'interval'
    lift_df.reset_index(inplace = True)
    lift_df.index = pd.Index(range(1, len(lift_df)+1), name='index')

    return lift_df


#%%

def get_coverage_by_group(
        data_df: pd.DataFrame,
        group_col = 'month',
        x_col = 'x',
    ):
    '''
    '''
    total_coverage = data_df[x_col].notna().mean()

    gp = data_df.groupby(by = group_col)
    sizes = gp.size()
    coverages = gp.apply(lambda df: df[x_col].notna().mean())
    stat_df = pd.DataFrame({
        'size': sizes, 'coverage': coverages
    })
    stat_df.loc['total'] = pd.Series({
        'size': len(data_df),
        'coverage': total_coverage,
    })
    stat_df['size'] = stat_df['size'].astype(int)
    return stat_df


#%%

def stat_y_by_month(
        data_df: pd.DataFrame,
        y_col: str = 'y',
        time_col: str = 'month',
        total = True,
        ):
    notna_idx = data_df[y_col].notna()
    y_df = data_df.loc[notna_idx, :]
    gp = y_df.groupby(by=time_col)[y_col]
    y_stat_df = pd.DataFrame({
        'count': gp.count(),
        'percent': gp.count() / len(y_df),
        'pos_cnt': gp.sum(),
        'pos_pct': gp.mean(),
    })
    if total:
        y_total_df = pd.DataFrame({
            'count': [y_stat_df['count'].sum()],
            'percent': [y_stat_df['percent'].sum()],
            'pos_cnt': [y_stat_df['pos_cnt'].sum()],
            'pos_pct': [y_stat_df['pos_cnt'].sum() / y_stat_df['count'].sum()],
        }, index=['total'])
        y_stat_df = pd.concat([y_stat_df, y_total_df], axis=0)
    y_stat_df.index.name = time_col
    y_stat_df.reset_index(inplace = True)
    y_stat_df.index = pd.Index(range(1, len(y_stat_df)+1), name='index')

    return y_stat_df


